name = 'serial_json'
version = '1.1.0'
description = 'Custom json serialization support.'
url = 'https://github.com/justengel/serial_json'
author = 'Justin Engel'
author_email = 'jtengel08@gmail.com'
